/*
 * FILE NAME
 *	arch/mips/vr41xx/vr4122/common/bcu.c
 *
 * BRIEF MODULE DESCRIPTION
 *	Bus Control Unit routines for the NEC VR4122 and VR4131.
 *
 * Author: Yoichi Yuasa
 *         yyuasa@mvista.com or source@mvista.com
 *
 * Copyright 2002 MontaVista Software Inc.
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 *  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 * Changes:
 *  Paul Mundt <lethal@chaoticdreams.org>
 *  - Calculate mips_counter_frequency properly on VR4131.
 *
 *  MontaVista Software Inc. <yyuasa@mvista.com> or <source@mvista.com>
 *  - New creation, NEC VR4122 and VR4131 are supported.
 */
#include <linux/types.h>

#include <asm/io.h>
#include <asm/param.h>
#include <asm/time.h>

#define CLKSPEEDREG			KSEG1ADDR(0x0f000014)
#define PMUTCLKDIVREG			KSEG1ADDR(0x0f0000cc)
#define calc_pclock(clksp)		((18432000 * 98) / ((clksp) & 0x001f))
#define calc_tclock(clksp,vtclock)	((vtclock) / (2 << (((clksp) & 0x1000) >> 12)))

unsigned long vr41xx_vtclock = 0;

static inline unsigned long calc_vtclock(u16 clksp, unsigned long pclock)
{
#ifdef CONFIG_VR4122_CLKSPEEDREG_FIX
	if ((clksp & 0x0700) == 0x0100) {
		u16 div;
		div = readw(PMUTCLKDIVREG);
		if ((div & 0x0007) == 0x0002)
			clksp = (clksp & ~0x0700) | 0x0200;
	}
#endif
	return pclock / ((clksp & 0x0700) >> 8);
}

void vr41xx_bcu_init(void)
{
	unsigned long pclock, tclock;
	u16 clksp;

	clksp = readw(CLKSPEEDREG);
	pclock = calc_pclock(clksp);
	printk(KERN_INFO "PClock: %ldHz\n", pclock);

	vr41xx_vtclock = calc_vtclock(clksp, pclock);
	printk(KERN_INFO "VTClock: %ldHz\n", vr41xx_vtclock);

	tclock = calc_tclock(clksp, vr41xx_vtclock);
	printk(KERN_INFO "TClock: %ldHz\n", tclock);

	/*
	 * The base clock divisor varies depending on which revision of the
	 * Vr4131 we happen to be dealing with ..
	 */
#ifdef CONFIG_VR4131
	{
		int prid = read_32bit_cp0_register(CP0_PRID);

		/*
		 * Revisions 2.0 and 2.1 use a value of (tclock / 2).
		 */
		if (prid == 0x0c81 || prid == 0x0c82)
			mips_counter_frequency = tclock / 2;
	}
#endif

	/*
	 * All other revisions of the Vr4131, as well as the Vr4122 use a
	 * value of (tclock / 4).
	 */
	if (!mips_counter_frequency)
		mips_counter_frequency = tclock / 4;
}
